from ..._helpers import isochoric_volumetric_split
from ....math import transpose, sum1, diag, sqrt, inv, det
from ..quadrature._bazant_oh import BazantOh
from .._chain import langevin, linear


@isochoric_volumetric_split
def microsphere_nonaffine_stretch(F, p, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Non-affine stretch part."

    r = quadrature.points
    w = quadrature.weights

    C = transpose(F) @ F
    nonaffine_stretch = sum1(sqrt(diag(r.T @ C @ r)) ** p * w) ** (1 / p)

    return f(nonaffine_stretch, **kwargs)


@isochoric_volumetric_split
def microsphere_nonaffine_tube(F, q, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Non-affine tube part."

    r = quadrature.points
    w = quadrature.weights

    Fs = det(F) * transpose(inv(F))
    Cs = transpose(Fs) @ Fs
    nonaffine_tubecontraction = sum1(sqrt(diag(r.T @ Cs @ r)) ** q * w)
    # nonaffine_areastretch = nonaffine_tube_contraction ** (1 / q)

    return f(nonaffine_tubecontraction, **kwargs)


@isochoric_volumetric_split
def microsphere_nonaffine_miehe_goektepe_lulei(F, mu, N, U, p, q):
    """Micro-sphere model: Combined non-affine stretch and
    tube model (for details see Miehe, Goektepe and Lulei (2004))."""

    kwargs_stretch = {"mu": mu, "N": N}
    kwargs_tube = {"mu": mu * N * U}

    quad = BazantOh(n=21)

    return microsphere_nonaffine_stretch(
        F, p=p, f=langevin, kwargs=kwargs_stretch, quadrature=quad
    ) + microsphere_nonaffine_tube(
        F, q=q, f=linear, kwargs=kwargs_tube, quadrature=quad
    )



import numpy as np
import matplotlib.pyplot as plt
from collections import namedtuple
from scipy.optimize import root


class LabCompressible:
    def __init__(self, material):

        self.material = material
        self.title = self._get_title()

    def _get_title(self):
        return "Material Formulation: " + "-".join(
            [m.title() for m in self.material.fun.__name__.split("_")]
        )

    def _uniaxial(self, stretch):
        "Load case `uniaxial`."

        def stress(stretch, stretch_2, stretch_3):
            "Evaluate the first Piola-Kirchhoff stress tensor."
            F = np.diag([stretch, stretch_2, stretch_3])
            return self.material.gradient([F])[0]

        def stability(stretch, stretches_23, stretches_23_eps):
            F = np.diag([stretch, *stretches_23])
            G = np.diag([stretch + 1e-6, *stretches_23_eps])
            A = self.material.hessian([F])[0]

            # convert hessian to (3, 3) matrix
            B = np.zeros((3, 3))
            c = [(0, 0), (1, 1), (2, 2)]

            for i, a in enumerate(c):
                for j, b in enumerate(c):
                    B[i, j] = A[(*a, *b)]

            # init unit force in direction 1
            df = np.zeros(3)
            df[0] = 1

            # calculate linear solution of stretch 1 resulting from unit load
            dl = (np.linalg.inv(B) @ df)[0]

            # check volume ratio
            J = stretch * stretches_23[0] * stretches_23[1]

            # check slope of force
            Q = self.material.gradient([G])[0][0, 0]
            P = self.material.gradient([F])[0][0, 0]

            return True if dl > 0 and J > 0 and (Q - P) > 0 else False

        def stress_free(stretches_23, stretch):
            s = stress(stretch, *stretches_23)
            return [s[1, 1], s[2, 2]]

        res = root(stress_free, np.ones(2), args=(stretch,))
        if not res.success:
            res = root(stress_free, np.ones(2) * 1 / np.sqrt(stretch), args=(stretch,))

        res_eps = root(stress_free, np.ones(2), args=(stretch + 1e-6,))
        if not res_eps.success:
            res_eps = root(
                stress_free,
                np.ones(2) * 1 / np.sqrt(stretch + 1e-6),
                args=(stretch + 1e-6,),
            )

        return (
            stress(stretch, *res.x)[0, 0],
            *res.x,
            stability(stretch, res.x, res_eps.x),
        )

    def _biaxial(self, stretch):
        "Load case `equi-biaxial`."

        def stress(stretch, stretch_3):
            "Evaluate the first Piola-Kirchhoff stress tensor."
            F = np.diag([stretch, stretch, stretch_3])
            return self.material.gradient([F])[0]

        def stability(stretch, stretch_3, stretch_3_eps):
            F = np.diag([stretch, stretch, stretch_3])
            G = np.diag([stretch + 1e-6, stretch + 1e-6, stretch_3_eps])
            A = self.material.hessian([F])[0]

            # convert hessian to (3, 3) matrix
            B = np.zeros((3, 3))
            c = [(0, 0), (1, 1), (2, 2)]

            for i, a in enumerate(c):
                for j, b in enumerate(c):
                    B[i, j] = A[(*a, *b)]

            # init unit force in direction 1
            df = np.zeros(3)
            df[0] = 1

            # calculate linear solution of stretch 1 resulting from unit load
            dl = (np.linalg.inv(B) @ df)[0]

            # check volume ratio
            J = stretch**2 * stretch_3

            # check slope of force
            Q = self.material.gradient([G])[0][0, 0]
            P = self.material.gradient([F])[0][0, 0]

            return True if dl > 0 and J > 0 and (Q - P) > 0 else False

        def stress_free(stretch_3, stretch):
            return [stress(stretch, *stretch_3)[2, 2]]

        res = root(stress_free, np.ones(1), args=(stretch,))
        stretch_3 = res.x[0]

        res_eps = root(stress_free, np.ones(1), args=(stretch + 1e-6,))
        stretch_3_eps = res_eps.x[0]

        return (
            stress(stretch, stretch_3)[0, 0],
            stretch,
            stretch_3,
            stability(stretch, stretch_3, stretch_3_eps),
        )

    def _planar(self, stretch):
        "Load case `planar shear`."

        def stress(stretch, stretch_3):
            "Evaluate the first Piola-Kirchhoff stress tensor."
            F = np.diag([stretch, 1, stretch_3])
            return self.material.gradient([F])[0]

        def stability(stretch, stretch_3, stretch_3_eps):
            F = np.diag([stretch, 1, stretch_3])
            G = np.diag([stretch + 1e-6, 1, stretch_3_eps])
            A = self.material.hessian([F])[0]

            # convert hessian to (3, 3) matrix
            B = np.zeros((3, 3))
            c = [(0, 0), (1, 1), (2, 2)]

            for i, a in enumerate(c):
                for j, b in enumerate(c):
                    B[i, j] = A[(*a, *b)]

            # init unit force in direction 1
            df = np.zeros(3)
            df[0] = 1

            # calculate linear solution of stretch 1 resulting from unit load
            dl = (np.linalg.inv(B) @ df)[0]

            # check volume ratio
            J = stretch * stretch_3

            # check slope of force
            Q = self.material.gradient([G])[0][0, 0]
            P = self.material.gradient([F])[0][0, 0]

            return True if dl > 0 and J > 0 and (Q - P) > 0 else False

        def stress_free(stretch_3, stretch):
            return [stress(stretch, *stretch_3)[2, 2]]

        res = root(stress_free, np.ones(1), args=(stretch,))
        stretch_3 = res.x[0]

        res_eps = root(stress_free, np.ones(1), args=(stretch + 1e-6,))
        stretch_3_eps = res_eps.x[0]

        return (
            stress(stretch, stretch_3)[0, 0],
            1,
            stretch_3,
            stability(stretch, stretch_3, stretch_3_eps),
        )

    def _shear(self, shear):
        "Load case `Shear`."

        def stress(shear, stretch_2, stretch_3):
            F = np.diag([1, stretch_2, stretch_3])
            F[0, 1] = shear
            return self.material.gradient([F])[0]

        def stress_free(stretches, shear):
            return [stress(shear, *stretches)[1, 1], stress(shear, *stretches)[2, 2]]

        res = root(stress_free, np.ones(2), args=(shear,))
        stretches = res.x

        return (
            stress(shear, *stretches)[0, 1],
            stretches[0],
            stretches[1],
            None,
        )

    def run(
        self,
        ux=True,
        bx=True,
        ps=True,
        shear=True,
        stretch_min=None,
        stretch_max=2.5,
        shear_max=1.0,
        num=50,
    ):
        "Run load cases `UX/BX/PS/Simple-Shear (Incompressible)`."

        out = []

        Data = namedtuple(
            "Data", "label stretch stretch_2 stretch_3 shear stress stability"
        )

        if ux:
            if stretch_min is None:
                stretch_min = max(0, 1 - (stretch_max - 1) / 5)

            stretch_ux = np.linspace(stretch_min, stretch_max, num)
            stress_ux, stretch_2_ux, stretch_3_ux, stability_ux = np.array(
                [self._uniaxial(s11) for s11 in stretch_ux]
            ).T
            ux_data = Data(
                "uniaxial",
                stretch_ux,
                stretch_2_ux,
                stretch_3_ux,
                0,
                stress_ux,
                stability_ux,
            )
            out.append(ux_data)

        if bx:
            stretch_bx = np.linspace(1, (stretch_max - 1) / 2 + 1, num)
            stress_bx, stretch_2_bx, stretch_3_bx, stability_bx = np.array(
                [self._biaxial(s11) for s11 in stretch_bx]
            ).T
            bx_data = Data(
                "biaxial",
                stretch_bx,
                stretch_2_bx,
                stretch_3_bx,
                0,
                stress_bx,
                stability_bx,
            )
            out.append(bx_data)

        if ps:
            stretch_ps = np.linspace(1, stretch_max, num)
            stress_ps, stretch_2_ps, stretch_3_ps, stability_ps = np.array(
                [self._planar(s11) for s11 in stretch_ps]
            ).T
            ps_data = Data(
                "planar",
                stretch_ps,
                stretch_2_ps,
                stretch_3_ps,
                0,
                stress_ps,
                stability_ps,
            )
            out.append(ps_data)

        if shear:
            deformation_sh = np.linspace(0, shear_max, num)
            stress_shear, stretch_2_sh, stretch_3_sh, stability_sh = np.array(
                [self._shear(g) for g in deformation_sh]
            ).T
            shear_data = Data(
                "shear",
                1,
                stretch_2_sh,
                stretch_3_sh,
                deformation_sh,
                stress_shear,
                stability_sh,
            )
            out.append(shear_data)

        return out

    def plot(self, data, stability=False):
        "Plot results of UX/BX/PS load cases."

        fig, ax = plt.subplots()

        lineargs = {
            "uniaxial": {"color": "C0"},
            "biaxial": {"color": "C1"},
            "planar": {"color": "C2"},
        }

        if data[-1].label == "shear":
            data = data[:-1]

        for d in data:

            if stability:

                stable = np.array(d.stability, dtype=bool)

                unstable_idx = np.arange(len(d.stretch))[~stable]
                stable_idx = np.arange(len(d.stretch))[stable]

                stress_stable = d.stress.copy()
                stress_stable[unstable_idx] = np.nan

                stress_unstable = d.stress.copy()
                stress_unstable[stable_idx] = np.nan

                ax.plot(d.stretch, stress_stable, **lineargs[d.label], label=d.label)
                ax.plot(
                    d.stretch,
                    stress_unstable,
                    **lineargs[d.label],
                    linestyle="--",
                )

            else:

                ax.plot(d.stretch, d.stress, **lineargs[d.label], label=d.label)

        ax.grid()
        ax.set_title(self.title, fontsize=10)
        ax.set_xlabel(r"stretch $\lambda_1 \quad \longrightarrow$")
        ax.set_ylabel(r"force per undeformed area $P_{11} \quad \longrightarrow$")
        ax.legend()

        fig.tight_layout()

        return fig, ax

    def plot_shear(self, data):
        "Plot results of shear load case."

        fig, ax = plt.subplots()

        lineargs = {
            "shear": {"color": "C3"},
        }

        if data[-1].label == "shear":
            d = data[-1]
        else:
            raise TypeError("No shear data found.")

        ax.plot(d.shear, d.stress, **lineargs[d.label], label=d.label)

        ax.grid()
        ax.set_title(self.title, fontsize=10)
        ax.set_xlabel(r"shear deformation $F_{12}=\gamma \quad \longrightarrow$")
        ax.set_ylabel(r"force per undeformed area $P_{12} \quad \longrightarrow$")
        ax.legend()

        fig.tight_layout()

        return fig, ax



import numpy as np

from . import Material, MaterialTensor, Variable
from .math import det, horzcat, vertcat, zeros, gradient as grad, trace, eye, Function


class TwoFieldVariation:
    def __init__(self, material):
        self.material = material
        p = Variable("p", 1, 1)
        self.x = [self.material.x[0], p]
        self.W = Material(self.x, self._fun)
        self.function = self.W.function
        self.hessian = self.W.hessian
        self.gradient_vector_product = self.W.gradient_vector_product
        self.hessian_vector_product = self.W.hessian_vector_product

    def _fun(self, x):
        F, p = x[:2]
        J = det(F)
        W = self.material.fun(F, **self.material.kwargs)
        U = self.material.fun(J ** (1 / 3) * eye(3), **self.material.kwargs)
        dWdF = grad(W, F)
        dWdJ = trace(dWdF @ F.T) / J / 3
        d2WdJdJ = trace(grad(dWdJ, F) @ F.T) / J / 3
        return W - U + 1 / d2WdJdJ * (p * dWdJ - p**2 / 2)

    def gradient(self, x, *args, **kwargs):
        return [*self.W.gradient(x, *args, **kwargs), None]


class TwoFieldVariationPlaneStrain:
    def __init__(self, material):
        self.material = material
        p = Variable("p", 1, 1)
        self.x = [self.material.x[0], p]
        self.W = Material(self.x, self._fun)
        self.function = self.W.function
        self.hessian = self.W.hessian
        self.gradient_vector_product = self.W.gradient_vector_product
        self.hessian_vector_product = self.W.hessian_vector_product

    def _fun(self, x):
        F, p = x[:2]
        F = horzcat(vertcat(x[0], zeros(1, 2)), zeros(3, 1))
        F[2, 2] = 1  # fixed thickness ratio `h / H = 1`
        J = det(F)
        W = self.material.fun(F, **self.material.kwargs)
        U = self.material.fun(J ** (1 / 3) * eye(3), **self.material.kwargs)

        f = Variable("f", 3, 3)
        w = self.material.fun(f, **self.material.kwargs)

        dwdf = grad(w, f)
        dwdj = trace(dwdf @ f.T) / det(f) / 3
        d2wdjdj = trace(grad(dwdj, f) @ f.T) / det(f) / 3

        dWdJ = Function("w", [f], [dwdj])(F)
        d2WdJdJ = Function("w", [f], [d2wdjdj])(F)

        return W - U + 1 / d2WdJdJ * (p * dWdJ - p**2 / 2)

    def gradient(self, x, *args, **kwargs):
        return [*self.W.gradient(x, *args, **kwargs), None]


class ThreeFieldVariation:
    def __init__(self, material):
        self.material = material
        p = Variable("p", 1, 1)
        J = Variable("J", 1, 1)
        self.x = [self.material.x[0], p, J]
        self.W = Material(self.x, self._fun)
        self.function = self.W.function
        self.hessian = self.W.hessian
        self.gradient_vector_product = self.W.gradient_vector_product
        self.hessian_vector_product = self.W.hessian_vector_product

    def _fun(self, x):
        F, p, J = x[:3]
        detF = det(F)
        Fmod = (J / detF) ** (1 / 3) * F
        return self.material.fun(Fmod, **self.material.kwargs) + p * (detF - J)

    def gradient(self, x, *args, **kwargs):
        return [*self.W.gradient(x, *args, **kwargs), None]


class ThreeFieldVariationPlaneStrain:
    def __init__(self, material):
        self.material = material
        p = Variable("p", 1, 1)
        J = Variable("J", 1, 1)
        self.x = [self.material.x[0], p, J]
        self.W = Material(self.x, self._fun)
        self.function = self.W.function
        self.hessian = self.W.hessian
        self.gradient_vector_product = self.W.gradient_vector_product
        self.hessian_vector_product = self.W.hessian_vector_product

    def _fun(self, x):
        F, p, J = x[:3]
        F = horzcat(vertcat(x[0], zeros(1, 2)), zeros(3, 1))
        F[2, 2] = 1  # fixed thickness ratio `h / H = 1`
        detF = det(F)
        Fmod = (J / detF) ** (1 / 3) * F
        return self.material.fun(Fmod, **self.material.kwargs) + p * (detF - J)

    def gradient(self, x, *args, **kwargs):
        return [*self.W.gradient(x, *args, **kwargs), None]


class MaterialHyperelastic:
    def __init__(self, fun, **kwargs):
        F = Variable("F", 3, 3)
        self.x = [F]
        self.fun = fun
        self.kwargs = kwargs
        self.W = Material(self.x, self._fun_wrapper, kwargs=self.kwargs)
        self.gradient_vector_product = self.W.gradient_vector_product
        self.hessian_vector_product = self.W.hessian_vector_product

    def _fun_wrapper(self, x, **kwargs):
        return self.fun(x[0], **kwargs)

    def function(self, x, *args, **kwargs):
        return self.W.function(x[:1], *args, **kwargs)

    def gradient(self, x, *args, **kwargs):
        return [*self.W.gradient(x[:1], *args, **kwargs), None]

    def hessian(self, x, *args, **kwargs):
        return self.W.hessian(x[:1], *args, **kwargs)


class MaterialHyperelasticPlaneStrain:
    def __init__(self, fun, **kwargs):
        F = Variable("F", 2, 2)
        self.x = [F]
        self.fun = fun
        self.kwargs = kwargs
        self.W = Material(self.x, self._fun_wrapper, kwargs=self.kwargs)
        self.gradient_vector_product = self.W.gradient_vector_product
        self.hessian_vector_product = self.W.hessian_vector_product

    def _fun_wrapper(self, x, **kwargs):
        F = horzcat(vertcat(x[0], zeros(1, 2)), zeros(3, 1))
        F[2, 2] = 1  # fixed thickness ratio `h / H = 1`
        return self.fun(F, **kwargs)

    def function(self, x, *args, **kwargs):
        return self.W.function(x[:1], *args, **kwargs)

    def gradient(self, x, *args, **kwargs):
        return [*self.W.gradient(x[:1], *args, **kwargs), None]

    def hessian(self, x, *args, **kwargs):
        return self.W.hessian(x[:1], *args, **kwargs)


class MaterialHyperelasticPlaneStressIncompressible(MaterialHyperelasticPlaneStrain):
    def __init__(self, fun, **kwargs):
        super().__init__(fun, **kwargs)

    def _fun_wrapper(self, x, **kwargs):
        F = horzcat(vertcat(x[0], zeros(1, 2)), zeros(3, 1))
        F[2, 2] = 1 / det(x[0])  # thickness ratio `h / H = 1 / (a / A)`
        return self.fun(F, **kwargs)


class MaterialHyperelasticPlaneStressLinearElastic(MaterialHyperelasticPlaneStrain):
    def __init__(self, fun, **kwargs):
        super().__init__(fun, **kwargs)

    def _fun_wrapper(self, x, **kwargs):
        F = horzcat(vertcat(x[0], zeros(1, 2)), zeros(3, 1))
        # stress-free thickness ratio for linear elastic material
        # s_33 != 0 = 2 mu e_33 + lmbda (e_11 + e_22 + e_33)
        # e_33 = - (e_11 + e_22) * lmbda / (2 mu + lmbda)
        # F_33 = 1 + e_33
        F[2, 2] = 1 - (F[0, 0] + F[1, 1] - 2) * (
            kwargs["lmbda"] / (2 * kwargs["mu"] + kwargs["lmbda"])
        )
        return self.fun(F, **kwargs)


class MaterialComposite:
    def __init__(self, materials):
        "Composite Material as a sum of a list of hyperelastic materials."
        self.materials = materials
        self.fun = self.composite

        # get number of variables defined in the first material
        self._n = len(self.materials[0].x)

    def composite(self):
        "Dummy function for plot title."
        return

    def function(self, x, **kwargs):
        fun = [m.function(x[: self._n], **kwargs) for m in self.materials]
        return [np.sum([f[a] for f in fun], 0) for a in range(len(fun[0]))]

    def gradient(self, x, **kwargs):
        grad = [m.gradient(x[: self._n], **kwargs)[: len(x)] for m in self.materials]
        res = [np.sum([g[a] for g in grad], 0) for a in range(len(grad[0]))]
        return [*res, None]

    def hessian(self, x, **kwargs):
        hess = [m.hessian(x[: self._n], **kwargs) for m in self.materials]
        return [np.sum([h[a] for h in hess], 0) for a in range(len(hess[0]))]


class MaterialTensorGeneral(MaterialTensor):
    def __init__(self, fun, statevars_shape=(1, 1), x=None, triu=True, **kwargs):
        """A (first Piola-Kirchhoff stress) tensor-based material definition with
        state variables of a given shape."""

        if x is None:
            x = [Variable("F", 3, 3)]

        try:
            # displacement-pressure split
            x.append(fun.p)
        except:
            pass

        # add state variables
        x.append(Variable("z", *statevars_shape))

        super().__init__(x=x, fun=fun, triu=triu, statevars=1, kwargs=kwargs)



from multiprocessing import cpu_count

import numpy as np
import casadi as ca

from ._apply import apply
from . import Variable


class Function:
    def __init__(self, x, fun, args=(), kwargs={}, compress=False):

        self.x = x
        self._fun = fun

        self.args = args
        self.kwargs = kwargs

        # generate function
        f = self._fun(self.x, *self.args, **self.kwargs)

        # check if function is list or tuple
        if isinstance(f, list) or isinstance(f, tuple):
            self._f = f
        else:
            self._f = [f]

        # generate casADi function objects
        self._function = ca.Function("f", self.x, self._f)

        # generate indices
        self._idx_x = [y.shape for y in x]
        self._idx_function = [y.shape for y in self._f]

    def function(self, x, threads=cpu_count()):
        "Return the function."
        return apply(
            x,
            fun=self._function,
            x_shape=self._idx_x,
            fun_shape=self._idx_function,
            threads=threads,
        )


class FunctionTensor:
    def __init__(self, x, fun, args=(), kwargs={}, compress=False):

        self.x = x
        self._fun = fun

        self.args = args
        self.kwargs = kwargs

        # generate function
        f = self._fun(self.x, *self.args, **self.kwargs)

        # check if function is list or tuple
        if isinstance(f, list) or isinstance(f, tuple):
            self._f = f
        else:
            self._f = [f]

        # generate casADi function objects
        self._function = ca.Function("f", self.x, self._f)

        # generate indices
        self._idx_function = [y.shape for y in self._f]
        self._idx_x = self._idx_function[: len(self.x)]

    def function(self, x, threads=cpu_count()):
        "Return the function."
        return apply(
            x,
            fun=self._function,
            x_shape=self._idx_x,
            fun_shape=self._idx_function,
            threads=threads,
        )


class Material(Function):
    def __init__(
        self, x, fun, args=(), kwargs={}, compress=False, triu=True, statevars=0
    ):

        # init Function
        super().__init__(x=x, fun=fun, args=args, kwargs=kwargs)

        # no. of active variables
        n = len(self.x) - statevars
        self._idx_gradient = self._idx_x[:n]

        _h_diag = []
        _hvp_diag = []

        self._h = []
        self._g = []

        self._hvp = []
        self._gvp = []

        # generate vectors for gradient- and hessian-vector products
        self.v = [Variable("v%d" % a, *x.shape) for a, x in enumerate(self.x)]
        self.u = [Variable("u%d" % a, *x.shape) for a, x in enumerate(self.x)]

        # alias
        self.jacobian = self.gradient

        # generate list of diagonal hessian entries and gradients
        # (including vector-products)
        for x, v, u in zip(self.x[:n], self.v[:n], self.u[:n]):
            _h, _g = ca.hessian(self._f[0], x)
            _h_diag.append(_h)
            self._g.append(_g)

            _gvp = ca.jtimes(self._f[0], x, v)
            _hvp = ca.jtimes(_gvp, x, u)
            self._gvp.append(_gvp)
            _hvp_diag.append(_hvp)

        # generate upper-triangle of hessian (-vector-products)
        for i, (g, gvp) in enumerate(zip(self._g, self._gvp)):
            for j, (x, u) in enumerate(zip(self.x[:n], self.u[:n])):
                if triu and j >= i or not triu:
                    if i != j:
                        self._h.append(ca.jacobian(g, x))
                        self._hvp.append(ca.jtimes(gvp, x, u))
                    else:
                        self._h.append(_h_diag[i])
                        self._hvp.append(_hvp_diag[i])

        # generate casADi function objects
        self._gradient = ca.Function("g", self.x, self._g)
        self._hessian = ca.Function("h", self.x, self._h)
        self._gradient_vector_product = ca.Function(
            "gvp", [*self.x, *self.v], self._gvp
        )
        self._hessian_vector_product = ca.Function(
            "hvp", [*self.x, *self.v, *self.u], self._hvp
        )

        # generate indices
        self._idx_hessian = []

        if compress:
            for i in range(len(self._idx_function)):
                if np.all(np.array(self._idx_function[i]) == 1):
                    self._idx_function[i] = ()

            for i in range(len(self._idx_gradient)):
                if np.all(np.array(self._idx_gradient[i]) == 1):
                    self._idx_gradient[i] = ()

        for i in range(len(self._idx_gradient)):
            a = self._idx_gradient[i]

            for j in range(len(self._idx_gradient)):
                b = self._idx_gradient[j]

                if triu and j >= i or not triu:
                    self._idx_hessian.append((*a, *b))

    def gradient(self, x, threads=cpu_count()):
        "Return list of gradients."
        return apply(
            x,
            fun=self._gradient,
            x_shape=self._idx_gradient,
            fun_shape=self._idx_gradient,
            threads=threads,
        )

    def hessian(self, x, threads=cpu_count()):
        "Return upper-triangle entries of hessian."
        return apply(
            x,
            fun=self._hessian,
            x_shape=self._idx_gradient,
            fun_shape=self._idx_hessian,
            threads=threads,
        )

    def gradient_vector_product(self, x, v, threads=cpu_count()):
        "Return list of gradient-vector-products."
        return apply(
            [*x, *v],
            fun=self._gradient_vector_product,
            x_shape=self._idx_gradient,
            fun_shape=self._idx_function * len(self._gvp),
            threads=threads,
        )

    def hessian_vector_product(self, x, v, u, threads=cpu_count()):
        "Return list of hessian-vector-products."
        return apply(
            [*x, *v, *u],
            fun=self._hessian_vector_product,
            x_shape=self._idx_gradient,
            fun_shape=self._idx_function * len(self._hvp),
            threads=threads,
        )


class MaterialTensor(FunctionTensor):
    def __init__(
        self, x, fun, args=(), kwargs={}, compress=False, triu=True, statevars=0
    ):

        # init Function
        super().__init__(x=x, fun=fun, args=args, kwargs=kwargs)

        # no. of active variables
        n = len(self.x) - statevars

        # generate vector for gradient-vector-product
        self.v = [Variable("v%d" % a, *x.shape) for a, x in enumerate(self.x)]

        # generate gradient and gradient-vector-product
        self._g = [ca.jacobian(f, x) for x in self.x[:n] for f in self._f[:n]]
        self._gvp = [
            ca.jtimes(f, x, v)
            for x, v in zip(self.x[:n], self.v[:n])
            for f in self._f[:n]
        ]

        # alias
        self.jacobian = self.gradient

        # store only upper-triangle entries of gradients
        if triu:
            i, j = np.triu_indices(len(self.x[:n]))
            a = (
                np.arange(len(self.x[:n]) ** 2)
                .reshape(len(self.x[:n]), len(self.x[:n]))[i, j]
                .ravel()
            )
            self._g = [self._g[b] for b in a]
            self._gvp = [self._gvp[b] for b in a]

        # generate casADi function objects
        self._function = ca.Function("f", self.x, self._f)
        self._gradient = ca.Function("g", self.x, self._g)
        self._gradient_vector_product = ca.Function(
            "gvp", [*self.x, *self.v], self._gvp
        )

        # generate indices
        self._idx_gradient = []

        if compress:
            for i in range(len(self._idx_function)):
                if np.all(np.array(self._idx_function[i]) == 1):
                    self._idx_x[i] = ()
                    self._idx_function[i] = ()

        for i in range(len(self._idx_function[:n])):
            a = self._idx_function[i]

            for j in range(len(self._idx_function[:n])):
                b = self._idx_function[j]

                if triu:
                    if j >= i:
                        self._idx_gradient.append((*a, *b))
                else:
                    self._idx_gradient.append((*a, *b))

    def gradient(self, x, threads=cpu_count()):
        "Return list of gradients."
        return apply(
            x,
            fun=self._gradient,
            x_shape=self._idx_x,
            fun_shape=self._idx_gradient,
            threads=threads,
        )

    def gradient_vector_product(self, x, threads=cpu_count()):
        "Return list of gradient-vector-products."
        return apply(
            x,
            fun=self._gradient_vector_product,
            x_shape=self._idx_x,
            fun_shape=self._idx_function * len(self._gvp),
            threads=threads,
        )



import numpy as np

from matadi import MaterialTensorGeneral
from matadi.models import NeoHookeOgdenRoxburgh, Morph, neo_hooke, volumetric
from matadi.math import det, gradient, ones_like, zeros_like


def fun(x, C10=0.5, bulk=5000):

    F = x[0]
    J = det(F)

    W = neo_hooke(F, C10)
    U = volumetric(J, bulk)

    statevars_old = x[-1]
    statevars_new = ones_like(statevars_old)  # only for testing
    statevars_new = zeros_like(statevars_old)

    return gradient(W, F) + gradient(U, F), statevars_new


def test_u_templates():

    Custom = MaterialTensorGeneral(fun, statevars_shape=(1, 1))

    # Material as a function of `F` and `p`
    # with additional state variables `z`
    for M in [Custom]:

        FF = (np.random.rand(3, 3, 8, 100) - 0.5) / 2
        zz = np.random.rand(*M.x[-1].shape, 8, 100)

        for a in range(3):
            FF[a, a] += 1

        P = M.function([FF, zz])  # stress, constraint, statevars_new
        A = M.gradient([FF, zz])

        assert len(P) == 2
        assert len(A) == 1


def test_up_templates():

    # Material as a function of `F` and `p`
    # with additional state variables `z`
    for M in [NeoHookeOgdenRoxburgh(), Morph()]:

        FF = (np.random.rand(3, 3, 8, 100) - 0.5) / 2
        pp = np.random.rand(1, 8, 100)
        zz = np.random.rand(*M.x[-1].shape, 8, 100)

        for a in range(3):
            FF[a, a] += 1

        P = M.function([FF, pp, zz])  # stress, constraint, statevars_new
        A = M.gradient([FF, pp, zz])

        assert len(P) == 3
        assert len(A) == 3


if __name__ == "__main__":
    test_u_templates()
    test_up_templates()



import numpy as np


def apply(x, fun, x_shape, fun_shape, threads=1):
    "Helper function for the calculation of fun(x)."

    # get shape of trailing axes
    trailing_axes = [len(y.shape) - len(y_shape) for y, y_shape in zip(x, x_shape)][0]

    if trailing_axes == 0:
        ax = [1]
    else:
        ax = x[0].shape[-trailing_axes:]

    def rshape(z):
        "Reshape array `z`: 'i,j,...->i,...'."
        if len(z.shape) == trailing_axes:
            return z.reshape(1, -1, order="F")
        else:
            return z.reshape(z.shape[0], -1, order="F")

    # apply reshape on input
    y = [rshape(z) for z in x]

    # threads dict
    if threads > 1:
        parallel = ("thread", threads)
    else:
        parallel = ()

    # map function `N` times on reshaped input
    N = np.product(ax)
    out = fun.map(N, *parallel)(*y)

    if not isinstance(out, tuple):
        out = (out,)

    # return 'i,j,...' reshaped output
    if trailing_axes == 0:
        ax = ()
        if fun_shape == [()]:
            fun_shape = [(1,)]

    return [np.array(o).reshape(*f, *ax, order="F") for o, f in zip(out, fun_shape)]



import numpy as np

from matadi import Variable, Material
from matadi.math import det, transpose, trace


def test_scalar():

    # variables
    F = Variable("F", 3, 3)
    p = Variable("p", 1)
    J = Variable("J", 1)

    def neohooke(x, mu=1.0, bulk=200.0):
        "Strain energy function of nearly-incompressible Neo-Hookean material."

        F, p, theta = x
        J = det(F)
        C = transpose(F) @ F
        I1 = J ** (-2 / 3) * trace(C)

        return mu * (I1 - 3) + bulk * (theta - 1) ** 2 / 2 + p * (J - theta)

    # data
    FF = np.random.rand(3, 3, 5, 100)
    pp = np.random.rand(5, 100)
    JJ = np.random.rand(5, 100)

    # functional
    W = Material(
        x=[F, p, J], fun=neohooke, kwargs={"mu": 1.0, "bulk": 10.0}, compress=False
    )

    dW = W.gradient([FF, pp, JJ])
    dW2 = W.jacobian([FF, pp, JJ])
    DW = W.hessian([FF, pp, JJ])

    assert dW[0].shape == (3, 3, 5, 100)
    assert dW[1].shape == (1, 1, 5, 100)
    assert dW[2].shape == (1, 1, 5, 100)

    assert dW2[2].shape == dW2[2].shape

    assert DW[0].shape == (3, 3, 3, 3, 5, 100)
    assert DW[1].shape == (3, 3, 1, 1, 5, 100)
    assert DW[2].shape == (3, 3, 1, 1, 5, 100)
    assert DW[3].shape == (1, 1, 1, 1, 5, 100)
    assert DW[4].shape == (1, 1, 1, 1, 5, 100)
    assert DW[5].shape == (1, 1, 1, 1, 5, 100)

    # dataset 2
    FF = np.random.rand(3, 3, 5, 100)
    pp = np.random.rand(1, 5, 100)
    JJ = np.random.rand(1, 5, 100)

    # functional
    W = Material(
        x=[F, p, J], fun=neohooke, kwargs={"mu": 1.0, "bulk": 10.0}, compress=False
    )

    dW = W.gradient([FF, pp, JJ])
    DW = W.hessian([FF, pp, JJ])


def test_scalar_compress():

    # variables
    F = Variable("F", 3, 3)
    p = Variable("p", 1)
    J = Variable("J", 1)

    def neohooke(x, mu=1.0, bulk=200.0):
        "Strain energy function of nearly-incompressible Neo-Hookean material."

        F, p, theta = x
        J = det(F)
        C = transpose(F) @ F
        I1 = J ** (-2 / 3) * trace(C)

        return mu * (I1 - 3) + bulk * (theta - 1) ** 2 / 2 + p * (J - theta)

    # data
    FF = np.random.rand(3, 3, 5, 100)
    pp = np.random.rand(5, 100)
    JJ = np.random.rand(5, 100)

    # functional
    W = Material(
        x=[F, p, J], fun=neohooke, kwargs={"mu": 1.0, "bulk": 10.0}, compress=True
    )

    dW = W.gradient([FF, pp, JJ])
    DW = W.hessian([FF, pp, JJ])

    assert dW[0].shape == (3, 3, 5, 100)
    assert dW[1].shape == (5, 100)
    assert dW[2].shape == (5, 100)

    assert DW[0].shape == (3, 3, 3, 3, 5, 100)
    assert DW[1].shape == (3, 3, 5, 100)
    assert DW[2].shape == (3, 3, 5, 100)
    assert DW[3].shape == (5, 100)
    assert DW[4].shape == (5, 100)
    assert DW[5].shape == (5, 100)

    # dataset 2
    FF = np.random.rand(3, 3, 5, 100)
    pp = np.random.rand(1, 5, 100)
    JJ = np.random.rand(1, 5, 100)

    # functional
    W = Material(
        x=[F, p, J], fun=neohooke, kwargs={"mu": 1.0, "bulk": 10.0}, compress=False
    )

    dW = W.gradient([FF, pp, JJ])
    DW = W.hessian([FF, pp, JJ])


if __name__ == "__main__":
    test_scalar()
    test_scalar_compress()



from .__about__ import __version__

import casadi

Variable = casadi.SX.sym

from ._material import (
    Function,
    Function as FunctionScalar,
    FunctionTensor,
    Material,
    Material as MaterialScalar,
    MaterialTensor,
)
from . import models
from . import math
from ._templates import (
    TwoFieldVariation,
    TwoFieldVariationPlaneStrain,
    ThreeFieldVariation,
    ThreeFieldVariationPlaneStrain,
    MaterialHyperelastic,
    MaterialComposite,
    MaterialHyperelasticPlaneStrain,
    MaterialHyperelasticPlaneStressIncompressible,
    MaterialHyperelasticPlaneStressLinearElastic,
    MaterialTensorGeneral,
)
from ._lab_compressible import (
    LabCompressible,
    LabCompressible as Lab,
)
from ._lab_incompressible import (
    LabIncompressible,
)


__all__ = [
    "__version__",
]



from ._models import microsphere_nonaffine_stretch as stretch
from ._models import microsphere_nonaffine_tube as tube
from ._models import microsphere_nonaffine_miehe_goektepe_lulei as miehe_goektepe_lulei



from ._helpers import isochoric_volumetric_split
from ..math import dot, det, transpose, trace, eigvals, sum1, log, sqrt, eye, sym


def linear_elastic(F, mu, lmbda):
    strain = sym(F - eye(3))
    return mu * trace(strain @ strain) + lmbda / 2 * trace(strain) ** 2


def saint_venant_kirchhoff(F, mu, lmbda):
    C = dot(transpose(F), F)
    I1 = trace(C) / 2 - 3 / 2
    I2 = trace(C @ C) / 4 - trace(C) / 2 + 3 / 4
    return mu * I2 + lmbda * I1**2 / 2


@isochoric_volumetric_split
def neo_hooke(F, C10):
    C = transpose(F) @ F
    I1 = trace(C)
    return C10 * (I1 - 3)


@isochoric_volumetric_split
def mooney_rivlin(F, C10, C01):
    C = transpose(F) @ F
    I1 = trace(C)
    I2 = (trace(C) ** 2 - trace(C @ C)) / 2
    return C10 * (I1 - 3) + C01 * (I2 - 3)


@isochoric_volumetric_split
def yeoh(F, C10, C20, C30):
    J = det(F)
    C = transpose(F) @ F
    I1 = J ** (-2 / 3) * trace(C)
    return C10 * (I1 - 3) + C20 * (I1 - 3) ** 2 + C30 * (I1 - 3) ** 3


@isochoric_volumetric_split
def third_order_deformation(F, C10, C01, C11, C20, C30):
    C = transpose(F) @ F
    I1 = trace(C)
    I2 = (trace(C) ** 2 - trace(C @ C)) / 2
    return (
        C10 * (I1 - 3)
        + C01 * (I2 - 3)
        + C11 * (I1 - 3) * (I2 - 3)
        + C20 * (I1 - 3) ** 2
        + C30 * (I1 - 3) ** 3
    )


@isochoric_volumetric_split
def ogden(F, mu, alpha):
    C = transpose(F) @ F
    wC = eigvals(C)

    out = 0
    for m, a in zip(mu, alpha):
        wk = wC ** (a / 2)
        out += 2 * m / a**2 * (sum1(wk)[0, 0] - 3)

    return out


@isochoric_volumetric_split
def arruda_boyce(F, C1, limit):
    C = transpose(F) @ F
    I1 = trace(C)

    alpha = [1 / 2, 1 / 20, 11 / 1050, 19 / 7000, 519 / 673750]
    beta = 1 / limit**2

    out = 0
    for i, a in enumerate(alpha):
        j = i + 1
        out += a * beta ** (2 * j - 2) * (I1**j - 3**j)

    return C1 * out


@isochoric_volumetric_split
def extended_tube(F, Gc, delta, Ge, beta):
    C = transpose(F) @ F
    D = trace(C)
    wC = eigvals(C)
    g = (1 - delta**2) * (D - 3) / (1 - delta**2 * (D - 3))
    Wc = Gc / 2 * (g + log(1 - delta**2 * (D - 3)))
    We = 2 * Ge / beta**2 * sum1(wC ** (-beta / 2) - 1)
    return Wc + We


@isochoric_volumetric_split
def van_der_waals(F, mu, limit, a, beta):
    C = transpose(F) @ F
    I1 = trace(C)
    I2 = (trace(C) ** 2 - trace(C @ C)) / 2
    I = (1 - beta) * I1 + beta * I2
    eta = sqrt((I - 3) / (limit**2 - 3))
    return mu * (
        -(limit**2 - 3) * (log(1 - eta) + eta) - 2 / 3 * a * ((I - 3) / 2) ** (3 / 2)
    )



import numpy as np

from matadi import (
    MaterialHyperelasticPlaneStrain,
    MaterialHyperelasticPlaneStressIncompressible,
    MaterialHyperelasticPlaneStressLinearElastic,
    ThreeFieldVariationPlaneStrain,
    TwoFieldVariationPlaneStrain,
)
from matadi.models import neo_hooke, linear_elastic


def pre():

    FF = np.random.rand(2, 2, 8, 1000)
    for a in range(2):
        FF[a, a] += 1

    return FF


def pre_mixed():

    FF = np.random.rand(2, 2, 8, 1000)
    for a in range(2):
        FF[a, a] += 1

    pp = np.random.rand(1, 1, 8, 1000)
    JJ = np.random.rand(1, 1, 8, 1000) + 1

    return FF, pp, JJ


def test_plane_strain():

    # data
    FF = pre()

    # init Material
    W = MaterialHyperelasticPlaneStrain(
        fun=neo_hooke,
        C10=0.5,
    )

    W0 = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)


def test_plane_strain_mixed():

    # data
    FF, pp, JJ = pre_mixed()

    # init Material
    W = MaterialHyperelasticPlaneStrain(
        fun=neo_hooke,
        C10=0.5,
    )

    W_upJ = ThreeFieldVariationPlaneStrain(W)

    W0 = W_upJ.function([FF, pp, JJ])
    dW = W_upJ.gradient([FF, pp, JJ])
    DW = W_upJ.hessian([FF, pp, JJ])

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)

    W_up = TwoFieldVariationPlaneStrain(W)

    W0 = W_up.function([FF, pp])
    dW = W_up.gradient([FF, pp])
    DW = W_up.hessian([FF, pp])

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)


def test_plane_stress_incompr():

    # data
    FF = pre()

    # init Material
    W = MaterialHyperelasticPlaneStressIncompressible(
        fun=neo_hooke,
        C10=0.5,
    )

    W0 = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)


def test_plane_stress_linear():

    # data
    FF = pre()

    # init Material
    W = MaterialHyperelasticPlaneStressLinearElastic(
        fun=linear_elastic,
        mu=1.0,
        lmbda=200.0,
    )

    W0 = W.function([FF])
    dW = W.gradient([FF])
    DW = W.hessian([FF])

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)

    assert W0[0].shape == (1, 1, 8, 1000)
    assert dW[0].shape == (2, 2, 8, 1000)
    assert DW[0].shape == (2, 2, 2, 2, 8, 1000)


if __name__ == "__main__":
    test_plane_strain()
    test_plane_strain_mixed()
    test_plane_stress_incompr()
    test_plane_stress_linear()



import numpy as np


class BazantOh:
    def __init__(self, n: int = 21):
        """ "Points and weights of a numeric integration scheme on the surface
        of a sphere.

        Bazant, Z. P., & Oh, B. H. (1986). Efficient Numerical Integration on
        the Surface of a Sphere. ZAMM ‐ Journal of Applied Mathematics and
        Mechanics / Zeitschrift für Angewandte Mathematik und Mechanik, 66(1),
        37-49. https://doi.org/10.1002/zamm.19860660108

        """

        schemes = {
            21: self._scheme_21,
        }

        self.points, self.weights = schemes[n]()
        self.bases = np.einsum("i...,j...->...ij", self.points, self.points).reshape(
            -1, 9
        )

    def _scheme_21(self):
        "2x21-point scheme (degree 9, orthogonal symmetries)."

        a = np.sqrt(2) / 2
        b = 0.836095596749
        c = 0.387907304067

        points = np.array(
            [
                [0, 0, 1],
                [0, 1, 0],
                [1, 0, 0],
                [0, a, a],
                [0, -a, a],
                [a, 0, a],
                [-a, 0, a],
                [a, a, 0],
                [-a, a, 0],
                [b, c, c],
                [-b, c, c],
                [b, -c, c],
                [-b, -c, c],
                [c, b, c],
                [-c, b, c],
                [c, -b, c],
                [-c, -b, c],
                [c, c, b],
                [-c, c, b],
                [c, -c, b],
                [-c, -c, b],
            ]
        ).T

        w1 = 0.0265214244093
        w2 = 0.0199301476312
        w3 = 0.0250712367487

        weights = 2 * np.concatenate(
            [np.repeat(w1, 3), np.repeat(w2, 6), np.repeat(w3, 12)]
        )

        return points, weights



from ._hyperelasticity_isotropic import neo_hooke
from ..math import (
    det,
    inv,
    dev,
    sym,
    sqrt,
    if_else,
    vertsplit,
    vertcat,
    asvoigt,
    astensor,
    tresca,
    mexp,
    gradient,
)


def morph(x, p1, p2, p3, p4, p5, p6, p7, p8):
    "MORPH consitutive material formulation."

    # split input into the deformation gradient and the vector of state variables
    F, statevars = x[0], x[-1]

    # split state variables
    CTSn, Cn, SZn = vertsplit(statevars, [0, 1, 7, 13])
    Cn, SZn = astensor(Cn), astensor(SZn)

    # determimant of deformtion gradient
    J = det(F)

    # isochoric part of right Cauchy-Green deformation tensor
    C = F.T @ F
    CG = J ** (-2 / 3) * C

    # incremental right Cauchy-Green deformation tensor
    dC = C - Cn

    # (isochoric part of) lagrangian rate of deformation tensor
    L = dev(sym(dC @ inv(C))) @ CG

    # tresca invariants of (distortional part of) C and L
    CT = tresca(CG)
    LT = tresca(L)

    # maximum historical tresca invariant of (distortional part of) C
    CTS = if_else(CT > CTSn, CT, CTSn)

    # stable normalizations: L / LT and CT / CTS
    L_LT = if_else(LT > 0, L / LT, L)
    CT_CTS = if_else(CTS > 0, CT / CTS, CT)

    # MORPH deformation-dependent material parameters
    f = lambda x: 1 / sqrt(1 + x**2)
    a = p1 + p2 * f(p3 * CTS)
    b = p4 * f(p3 * CTS)
    c = p5 * CTS * (1 - f(CTS / p6))

    # Hull stress
    SH = (c * mexp(p7 * L_LT * CT_CTS) + p8 * L_LT) @ inv(C)

    # implict euler update of overstress evolution equation
    SZ = (SZn + b * LT * SH) / (1 + b * LT)

    # hyperelastic part of strain energy function
    W = neo_hooke(F, C10=0.5)

    # final overstress as first Piola-Kirchhoff stress
    PZ = F @ (dev(SZ @ C) @ inv(C))

    # update state variables
    statevars_new = vertcat(CTS, asvoigt(C), asvoigt(SZ))

    # total first Piola-Kirchhoff stress and new state variables
    return [2 * a * gradient(W, F) + PZ, statevars_new]



from ..math import fmax, erf


def ogden_roxburgh(W, Wmaxn, r, m, beta):
    """An isotropic pseudo-elastic material formulation for the description of
    the Mullins-softening of rubber-like materials according to
    Ogden & Roxburgh."""

    # update the maximum (isochoric part of the) strain energy density
    # of the total load history (state variable)
    Wmax = fmax(W, Wmaxn)

    # evaluate the relative softening `eta`
    eta = 1 - 1 / r * erf((Wmax - W) / (m + beta * Wmax))

    return eta, Wmax
